#ifndef _LOGHEAP_H_
#define _LOGHEAP_H_

#if USE_LOGHEAP

#include "pageinfo.h"
#include "memusage.h"

#define LOG_BUF_SZ		1000
#define LOG_ALLOCATION_INTERVAL	(32 * 1024)		// 64KB allocated bytes
#define LOG_PAGEACCESS_INTERVAL	512				// 256 pages accessed


//#define LOG_TRACE		1
#define LOG_USAGE		1
//#define LOG_REQUESTED	1


// choose at most one below
#define USE_ALLOCATION_TIME		1	// use allocated bytes as the scale of time
//#define USE_APPLICATION_TIME		1	// use (pre-measured) application time as the scale of time
//#define USE_PAGEACCESS_TIME		1	// use the number of heap pages accessed as the time scale
//#define USE_PAGELOCALITY_TIME	1	// use the events of protected pages being accessed as time ticks

#if (USE_ALLOCATION_TIME + USE_APPLICATION_TIME + USE_PAGEACCESS_TIME + USE_PAGELOCALITY_TIME) > 1
#error "hey, you can't choose more than one from USE_ALLOCATION_TIME, USE_APPLICATION_TIME, USE_PAGEACCESS_TIME and USE_PAGELOCALITY_TIME!"
#elif USE_PAGEACCESS_TIME && !TRACK_PAGE_ACCESS
#error "hey, you must also choose TRACK_PAGE_ACCESS when you choose USE_PAGEACCESS_TIME"
#elif USE_PAGELOCALITY_TIME && !TRACK_PAGE_ACCESS
#error "hey, you must also choose TRACK_PAGE_ACCESS when you choose USE_PAGELOCALITY_TIME"
#endif

#endif	// USE_LOGHEAP

namespace HL {

template<class SuperHeap>
class LogHeap : public SuperHeap {

#if USE_LOGHEAP
  public:
	LogHeap() : num_allocations(0), bytes_ever_requested(0), bytes_ever_allocated(0), objects_in_use(0), bytes_allocated_in_use(0), num_malloc_free(0) {
		dbprintf("LogHeap()\n");
#if LOG_REQUESTED
		bytes_requested_in_use = 0;
#endif	// LOG_REQUESTED
#if USE_APPLICATION_TIME
		next_stop = getNextStop();
#endif
	}

	~LogHeap() {
		dbprintf("~LogHeap()\n");
	}


	inline void * malloc(size_t size) {
		if (inside) {
			dbprintf("reentrant\n");
			assert(anyThreadCreated);
			assert(false);
		}
#if 0
		static int zero_malloc = 0;
		if (size == 0) {
			zero_malloc++;
			dbprintf("LogHeap::malloc(%d)...#%d\n", size, zero_malloc);
			return NULL;
		}
#else
		if (size == 0) {
			size = sizeof(double);
		}
#endif
		assert(size > 0);
//		assert(size % sizeof(double) == 0);

		size_t requested_size = size;

#if LOG_REQUESTED
		size += sizeof(double);
#endif

		inside = 1;
		void * ptr = SuperHeap::malloc(size);
		assert(ptr != NULL);
		memset(ptr, 0, size);
		//dbprintf("Leaving #%d malloc(%d)=%p\n", num_allocations, size, ptr);
		inside = 0;

		size_t allocated_size = SuperHeap::getSize(ptr);
		assert(allocated_size >= size);

		// update counts
		num_allocations++;
		bytes_ever_requested += requested_size;
		bytes_ever_allocated += allocated_size;
		objects_in_use++;
		bytes_allocated_in_use += allocated_size;
		num_malloc_free++;

		trace_printf("malloc %d %p %d\n", allocated_size, ptr, requested_size);

#if LOG_REQUESTED
		assert(allocated_size >= requested_size + sizeof(double));
		bytes_requested_in_use += requested_size;
		*(reinterpret_cast<size_t *>(ptr)) = requested_size;
		*(reinterpret_cast<size_t *>(ptr) + 1) = requested_size;
//		dbprintf("ptr=%p =%d =%d\n", ptr, *(reinterpret_cast<size_t *>(ptr)), *(reinterpret_cast<size_t *>(ptr) + 1));
		ptr = reinterpret_cast<void *>(reinterpret_cast<size_t>(ptr) + sizeof(double));
#endif	// LOG_REQUESTED

#if USE_ALLOCATION_TIME
		static long long last_logged = 0;
//		dbprintf("last_logged=%lld bytes_ever_requested=%lld requested_size=%d allocated_size=%d\n", last_logged, bytes_ever_requested, requested_size, allocated_size);
		assert(last_logged < bytes_ever_requested);
		if (bytes_ever_requested - last_logged >= LOG_ALLOCATION_INTERVAL) {
			OutputUsage();
			last_logged = bytes_ever_requested;
		}
#endif	//USE_ALLOCATION_TIME

#if USE_APPLICATION_TIME
		if (num_malloc_free == next_stop) {
			OutputUsage();
			next_stop = getNextStop();
		}
#endif	//USE_APPLICATION_TIME

#if USE_PAGEACCESS_TIME
		if (numAccessedPages >= LOG_PAGEACCESS_INTERVAL) {
			OutputUsage();
		}
#endif	//USE_PAGEACCESS_TIME

#if USE_PAGELOCALITY_TIME
		if (newPageAccessed) {
			newPageAccessed = false;
			OutputUsage();
		}
#endif	//USE_PAGEACCESS_TIME

		assert(currentMmapDiscarded <= currentAllocatedMmap);
		assert(currentSbrkDiscarded <= currentAllocatedSbrk);
		assert(bytes_allocated_in_use <= currentAllocatedMmap + currentAllocatedSbrk);
//		assert(bytes_allocated_in_use <= currentAllocatedMmap + currentAllocatedSbrk - currentMmapDiscarded - currentSbrkDiscarded);
		return ptr;
	}

	inline void free(void * ptr) {
//		assert(ptr != NULL);
#if 1
		static int null_free = 0;
		if (ptr == NULL) {
			null_free++;
			dbprintf("LogHeap::free(%p)...#%d\n", ptr, null_free);
			return;
		}
#endif

#if LOG_REQUESTED
		ptr = reinterpret_cast<void *>(reinterpret_cast<size_t>(ptr) - sizeof(double));
//		dbprintf("ptr=%p =%d =%d\n", ptr, *(reinterpret_cast<size_t *>(ptr)), *(reinterpret_cast<size_t *>(ptr) + 1));
		size_t requested_size = *(reinterpret_cast<size_t *>(ptr));
		assert(*(reinterpret_cast<size_t *>(ptr) + 1) == requested_size);
		bytes_requested_in_use -= requested_size;
#endif	// LOG_REQUESTED
		//dbprintf("[log] free %d in use %d\n", allocated_size, bytes_allocated_in_use);
		if (inside) {
			dbprintf("reentrant\n");
			assert(anyThreadCreated);
			assert(false);
		}
		inside = 1;

		size_t allocated_size = SuperHeap::getSize(ptr);

#if LOG_REQUESTED
		assert(allocated_size >= requested_size + sizeof(double));
#endif	// LOG_REQUESTED

		SuperHeap::free(ptr);
		//dbprintf("Leaving #%d free(%p)\n", num_free, ptr);
		inside = 0;

		// update counts
		objects_in_use--;
		bytes_allocated_in_use -= allocated_size;
		num_malloc_free++;

		trace_printf("free %d %p\n", allocated_size, ptr);

#if USE_APPLICATION_TIME
		if (num_malloc_free == next_stop) {
			OutputUsage();
			next_stop = getNextStop();
		}
#endif	//USE_APPLICATION_TIME

#if USE_PAGEACCESS_TIME
		if (numAccessedPages >= LOG_PAGEACCESS_INTERVAL) {
			OutputUsage();
		}
#endif	//USE_PAGEACCESS_TIME
	}

	inline void * realloc(void * ptr, size_t size) {
/*
		size_t sz = size;
    if (ptr == 0) {
      return malloc (sz);
    }
    if (sz == 0) {
      free (ptr);
      return 0;
    }

    size_t objSize = getSize (ptr);
    if (objSize == sz) {
      return ptr;
    }

    // Allocate a new block of size sz.
    void * buf = malloc (sz);

    // Copy the contents of the original object
    // up to the size of the new block.

    size_t minSize = (objSize < sz) ? objSize : sz;
    if (buf) {
      memcpy (buf, ptr, minSize);
    }

    // Free the old block.
    free (ptr);
    return buf;
*/

		assert(reinterpret_cast<size_t>(ptr) % sizeof(double) == 0);
		assert(size >= 0);

		if (inside) {
			dbprintf("reentrant\n");
			assert(anyThreadCreated);
			assert(false);
		}
		inside = 1;

		size_t requested_size;
		size_t allocated_size;

		if (ptr != NULL) {

#if LOG_REQUESTED
			ptr = reinterpret_cast<void *>(reinterpret_cast<size_t>(ptr) - sizeof(double));
//			dbprintf("ptr=%p =%d =%d\n", ptr, *(reinterpret_cast<size_t *>(ptr)), *(reinterpret_cast<size_t *>(ptr) + 1));
			requested_size = *(reinterpret_cast<size_t *>(ptr));
			assert(*(reinterpret_cast<size_t *>(ptr) + 1) == requested_size);
			bytes_requested_in_use -= requested_size;
#endif	// LOG_REQUESTED

			size_t allocated_size = SuperHeap::getSize(ptr);
			assert(allocated_size > 0 && allocated_size % sizeof(double ) == 0);

#if LOG_REQUESTED
			assert(allocated_size >= requested_size + sizeof(double));
#endif	// LOG_REQUESTED

			// update counts
			objects_in_use--;
			bytes_allocated_in_use -= allocated_size;
			num_malloc_free++;
		}

		bool need_alloc = false;

		if (size > 0) {
			need_alloc = true;
			requested_size = size;

#if LOG_REQUESTED
			size += sizeof(double);
#endif
		}

		void * newptr = SuperHeap::realloc(ptr, size);
		assert(size > 0 || newptr == NULL);
		allocated_size = 0;

		if (need_alloc) {
			size_t allocated_size = SuperHeap::getSize(newptr);
			assert(allocated_size >= size);

			// update counts
			num_allocations++;
			bytes_ever_requested += requested_size;
			bytes_ever_allocated += allocated_size;
			objects_in_use++;
			bytes_allocated_in_use += allocated_size;
			num_malloc_free++;

#if LOG_REQUESTED
			assert(allocated_size >= requested_size + sizeof(double));
			bytes_requested_in_use += requested_size;
			*(reinterpret_cast<size_t *>(newptr)) = requested_size;
			*(reinterpret_cast<size_t *>(newptr) + 1) = requested_size;
//			dbprintf("newptr=%p =%d =%d\n", newptr, *(reinterpret_cast<size_t *>(newptr)), *(reinterpret_cast<size_t *>(newptr) + 1));
			newptr = reinterpret_cast<void *>(reinterpret_cast<size_t>(newptr) + sizeof(double));
#endif	// LOG_REQUESTED
		}

		trace_printf("realloc %d %p %d %p\n", allocated_size, newptr, requested_size, ptr);

#if USE_ALLOCATION_TIME
		static long long last_logged = 0;
//		dbprintf("last_logged=%lld bytes_ever_requested=%lld requested_size=%d allocated_size=%d\n", last_logged, bytes_ever_requested, requested_size, allocated_size);
		assert(last_logged < bytes_ever_requested);
		if (bytes_ever_requested - last_logged >= LOG_ALLOCATION_INTERVAL) {
			OutputUsage();
			last_logged = bytes_ever_requested;
		}
#endif	//USE_ALLOCATION_TIME

#if USE_APPLICATION_TIME
		if (num_malloc_free == next_stop) {
			OutputUsage();
			next_stop = getNextStop();
		}
#endif	//USE_APPLICATION_TIME

#if USE_PAGEACCESS_TIME
		if (numAccessedPages >= LOG_PAGEACCESS_INTERVAL) {
			OutputUsage();
		}
#endif	//USE_PAGEACCESS_TIME

#if USE_PAGELOCALITY_TIME
		if (newPageAccessed) {
			newPageAccessed = false;
			OutputUsage();
		}
#endif	//USE_PAGEACCESS_TIME

		assert(currentMmapDiscarded <= currentAllocatedMmap);
		assert(currentSbrkDiscarded <= currentAllocatedSbrk);
		assert(bytes_allocated_in_use <= currentAllocatedMmap + currentAllocatedSbrk);
//		assert(bytes_allocated_in_use <= currentAllocatedMmap + currentAllocatedSbrk - currentMmapDiscarded - currentSbrkDiscarded);

		inside = 0;
		return newptr;
	}

#if LOG_REQUESTED
	inline size_t getSize(void * ptr) {
		ptr = reinterpret_cast<void *>(reinterpret_cast<size_t>(ptr) - sizeof(double));
		size_t size = SuperHeap::getSize(ptr);
		return size;
	}
#endif // LOG_REQUESTED

  private:
	static int inside;
/*
	static size_t num_allocations;
	static long long bytes_ever_allocated;
	static size_t objects_in_use;
	static size_t bytes_allocated_in_use;
*/
	size_t num_allocations;
	long long bytes_ever_requested;
	long long bytes_ever_allocated;
	size_t objects_in_use;
	size_t bytes_allocated_in_use;
	size_t num_malloc_free;
#if LOG_REQUESTED
	size_t bytes_requested_in_use;
#endif	// LOG_REQUESTED
#if USE_APPLICATION_TIME
	size_t next_stop;

	size_t getNextStop() {
		static bool initialized = false;
		static char buf[50000];
		static char * s;

		if (!initialized) {
			int ti_file = open("ti.log", O_RDONLY);
			if (ti_file < 0) {
				abort();
			}

			int rc = read(ti_file, buf, sizeof(buf));
			dbprintf("read() returns %d\n", rc);
			assert(rc > 0 && rc < sizeof(buf));

			s = buf;
			initialized = true;
		}

		char * p = strstr(s, "T");

		int tick = 0;
		if (p != NULL) {
			int rc = sscanf(p, "T %d", &tick);
			assert(rc == 1);
			s = p + 1;
		}

		dbprintf("tick=%d\n", tick);
		return tick;
	}
#endif	//USE_APPLICATION_TIME

	void trace_printf(const char * fmt, ...) {
#if LOG_TRACE
		static int trace_log = open("trace.log", O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
		if (trace_log < 0) {
			abort();
		}

		char buf[LOG_BUF_SZ];
		va_list ap;
		va_start(ap, fmt);

		int n = vsnprintf(buf, LOG_BUF_SZ, fmt, ap);
		if (n < 0 || n >= LOG_BUF_SZ) {
			abort();
		}

		write(trace_log, buf, n);
#endif
	}

	void usage_printf(const char * fmt, ...) {
#if LOG_USAGE
		static int usage_log = open("usage.log", O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR);
		if (usage_log < 0) {
			abort();
		}

		char buf[LOG_BUF_SZ];
		va_list ap;
		va_start(ap, fmt);

		int n = vsnprintf(buf, LOG_BUF_SZ, fmt, ap);
		if (n < 0 || n >= LOG_BUF_SZ) {
			abort();
		}

		write(usage_log, buf, n);
#endif
	}


	void OutputUsage() {
#if LOG_USAGE
		assert((currentMmapDiscarded & ~PAGE_MASK) == 0);
		assert((currentSbrkDiscarded & ~PAGE_MASK) == 0);

		if (bytes_allocated_in_use > currentAllocatedMmap + currentAllocatedSbrk - currentMmapDiscarded - currentSbrkDiscarded) {
			dbprintf("oops! see the following\n");
		}
//		dbprintf("num_allocations=%d bytes_ever_allocated=%lld objects_in_use=%d bytes_allocated_in_use=%d total_pages_mapped=%d total_pages_in_use=%d\n", num_allocations, bytes_ever_allocated, objects_in_use, bytes_allocated_in_use, (currentAllocatedMmap + currentAllocatedSbrk) >> PAGE_SHIFT, (currentAllocatedMmap + currentAllocatedSbrk - currentMmapDiscarded - currentSbrkDiscarded) >> PAGE_SHIFT);

//		dbprintf("currentAllocatedMmap=%d currentAllocatedSbrk=%d currentMmapDiscarded=%d currentSbrkDiscarded=%d\n", currentAllocatedMmap, currentAllocatedSbrk, currentMmapDiscarded, currentSbrkDiscarded);

//		dbprintf("currentAllocatedMmap+currentAllocatedSbrk-currentMmapDiscarded-currentSbrkDiscarded=%d -bytes_allocated_in_use=%d\n", currentAllocatedMmap + currentAllocatedSbrk - currentMmapDiscarded - currentSbrkDiscarded, currentAllocatedMmap + currentAllocatedSbrk - currentMmapDiscarded - currentSbrkDiscarded - bytes_allocated_in_use);

#if LOG_REQUESTED
		usage_printf("REQUESTED: num_allocations %d bytes_ever_requested %lld bytes_requested_in_use %d\n", num_allocations, bytes_ever_requested, bytes_requested_in_use);
#endif	// LOG_REQUESTED

		usage_printf("num_allocations %d bytes_ever_requested %lld bytes_ever_allocated %lld objects_in_use %d bytes_allocated_in_use %d pages_mapped %d pages_untouched %d pages_discarded %d pages_in_use %d\n", num_allocations, bytes_ever_requested, bytes_ever_allocated, objects_in_use, bytes_allocated_in_use, (currentAllocatedMmap + currentAllocatedSbrk) >> PAGE_SHIFT, (currentMmapUntouched + currentSbrkUntouched) >> PAGE_SHIFT, (currentMmapDiscarded + currentSbrkDiscarded) >> PAGE_SHIFT, (currentAllocatedMmap + currentAllocatedSbrk - currentMmapUntouched - currentSbrkUntouched - currentMmapDiscarded - currentSbrkDiscarded) >> PAGE_SHIFT);

#if USE_PAGEACCESS_TIME
		resetAllPages();
#endif	// USE_PAGEACCESS_TIME

#endif	// LOG_USAGE
	}

#endif	// USE_LOG_HEAP
};	//end of class LogHeap

#if USE_LOGHEAP

template<class SuperHeap>
int LogHeap<SuperHeap>::inside = 0;
/*
template<class SuperHeap>
size_t LogHeap<SuperHeap>::num_allocations = 0;

template<class SuperHeap>
long long LogHeap<SuperHeap>::bytes_ever_allocated = 0;

template<class SuperHeap>
size_t LogHeap<SuperHeap>::objects_in_use = 0;

template<class SuperHeap>
size_t LogHeap<SuperHeap>::bytes_allocated_in_use = 0;
*/
#endif	// USE_LOGHEAP

};	//end of namespace HL

#endif
